import{r as t,j as r}from"./@radix-DP6vWzyx.js";import{Q as i,W as o,j as l,aR as m}from"./index-QQb7wQEC.js";const c=e=>t.createElement("svg",{viewBox:"0 0 24 24",fill:"black",xmlns:"http://www.w3.org/2000/svg",...e},t.createElement("path",{d:"M12 8C12.5523 8 13 8.44771 13 9V14C13 14.5523 12.5523 15 12 15C11.4477 15 11 14.5523 11 14V9C11 8.44771 11.4477 8 12 8Z"}),t.createElement("path",{d:"M12 18C12.5523 18 13 17.5523 13 17C13 16.4477 12.5523 16 12 16C11.4477 16 11 16.4477 11 17C11 17.5523 11.4477 18 12 18Z"}),t.createElement("path",{fillRule:"evenodd",clipRule:"evenodd",d:"M10.2476 2.11111C11.0074 0.729566 12.9926 0.729561 13.7524 2.11111L23.0612 19.0361C23.7943 20.369 22.8299 21.9999 21.3087 21.9999H2.69127C1.17006 21.9999 0.205733 20.369 0.938833 19.0361L10.2476 2.11111ZM21.3087 19.9999L12 3.07495L2.69126 19.9999L21.3087 19.9999Z"})),d=m("flex items-center text-text-sm rounded-md border px-4 py-3",{variants:{intent:{primary:"border-primary-400 bg-primary-25",warning:"bg-[#FFF6EA] border-theme-border-moderate",neutral:"border-theme-border-moderate",error:"bg-error-50 border-error-300"}},defaultVariants:{intent:"primary"}});function h({children:e,className:n,intent:a,...s}){return r.jsxs("div",{...s,className:i(d({intent:a}),n),children:[r.jsx(x,{intent:a}),r.jsx("div",{className:"w-full min-w-0",children:e})]})}function x({intent:e}){switch(e){case"warning":return r.jsx(c,{className:"mr-4 h-5 w-5 shrink-0 fill-warning-700"});case"error":return r.jsx(l,{className:"mr-4 h-5 w-5 shrink-0 fill-error-700"});default:return r.jsx(o,{className:"mr-4 h-5 w-5 shrink-0 fill-theme-text-brand"})}}export{h as I};
