import{z as s}from"./index-QQb7wQEC.js";const e=s.string().min(8,"Password must be at least 8 characters").regex(/[A-Z]/,"Password must contain at least one uppercase letter").regex(/[a-z]/,"Password must contain at least one lowercase letter").regex(/[0-9]/,"Password must contain at least one number").regex(/[!@#$%^&*(),.?":{}|<>\/]/,"Password must contain at least one special character"),o=s.object({oldPassword:s.string().optional(),newPassword:e,confirmPassword:s.string().min(1)}),t=o.refine(a=>a.newPassword===a.confirmPassword,{path:["confirmPassword"],message:"Passwords do not match"});export{o as a,t as u};
