"""Release [0.37.0].

Revision ID: 0.37.0
Revises: b4eccf34dfa3
Create Date: 2023-04-03 18:04:47.776420

"""

# revision identifiers, used by Alembic.
revision = "0.37.0"
down_revision = "b4eccf34dfa3"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
