"""Release [0.54.0].

Revision ID: 0.54.0
Revises: 6917bce75069
Create Date: 2024-01-08 16:53:52.587158

"""

# revision identifiers, used by Alembic.
revision = "0.54.0"
down_revision = "6917bce75069"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
