import{j as e}from"./@radix-DP6vWzyx.js";import{A as s,b as a}from"./index-QQb7wQEC.js";function m({username:t}){return e.jsxs("div",{className:"inline-flex items-center gap-1",children:[e.jsx(s,{size:"sm",children:e.jsx(a,{size:"sm",children:t[0]})}),e.jsx("p",{className:"text-text-sm font-semibold text-theme-text-primary",children:t})]})}export{m as I};
