import{j as e}from"./@radix-DP6vWzyx.js";import{y as n}from"./index-QQb7wQEC.js";import{C as o}from"./ComponentIcon-CAIoUis2.js";function m({type:s,children:r}){return e.jsxs(n,{rounded:!1,className:"inline-flex shrink items-center gap-0.5 overflow-x-hidden text-text-sm",color:"purple",emphasis:"minimal",children:[e.jsx(o,{type:s,className:"h-4 w-4 shrink-0 fill-current"}),e.jsx("div",{className:"truncate",children:r})]})}export{m as C};
