import{r as p,j as e}from"./@radix-DP6vWzyx.js";import{a0 as m,a1 as c,a2 as d,a3 as x,Q as b}from"./index-QQb7wQEC.js";import{S as f}from"./chevron-down-8wLBS5pQ.js";function g({title:r,children:s,initialOpen:o=!1,className:l,contentClassName:t,intent:n="primary"}){const[a,i]=p.useState(o);return e.jsxs(m,{className:l,open:a,onOpenChange:i,children:[e.jsx(c,{intent:n,children:e.jsxs(d,{className:"flex w-full items-center gap-[10px]",children:[e.jsx(f,{className:` ${a?"":"-rotate-90"} h-5 w-5 rounded-md fill-neutral-500 transition-transform duration-200 hover:bg-neutral-200`}),r]})}),e.jsx(x,{className:b("space-y-3 border-t border-theme-border-moderate bg-theme-surface-primary px-5 py-3",t),children:s})]})}export{g as C};
