import{j as e}from"./@radix-DP6vWzyx.js";import{u as i,A as o,a as c,g as d,b as x,S as a,c as m,d as u,B as n,r as f,e as h,f as p}from"./index-QQb7wQEC.js";import{C as j}from"./CodeSnippet-CqybNv0k.js";import{S as g}from"./link-external-CBEk6kEG.js";import{L as v}from"./@react-router-BMhZulnd.js";import"./@tanstack-BUCbhJyH.js";import"./@reactflow-8U9qNlMR.js";import"./copy-CaSMXwiU.js";const w="/assets/cloud-squares-DeRLMopf.svg",b="/assets/codespaces-BitYDX9d.gif";function N(){var r,t,l;const{data:s}=i({throwOnError:!0});return e.jsx("div",{className:"w-full items-center border-b border-theme-border-moderate bg-white py-5",children:e.jsx("div",{className:"layout-container flex flex-col flex-wrap items-center justify-between md:flex-row",children:e.jsxs("div",{className:"flex flex-col items-center gap-5 md:flex-row",children:[s?e.jsxs(o,{size:"xxl",type:"square",children:[e.jsx(c,{alt:"Logo displayed for the server",src:d(((r=s.body)==null?void 0:r.server_name)||"default",64)}),e.jsx(x,{size:"xxl",children:((t=s.body)==null?void 0:t.server_name[0])||"D"})]}):e.jsx(a,{className:"h-[96px] w-[96px]"}),e.jsx("div",{children:e.jsxs("div",{className:"flex flex-col gap-2 md:flex-row",children:[s?e.jsx("h1",{className:"mb-1 text-display-xs font-semibold",children:((l=s.body)==null?void 0:l.server_name)||"default"}):e.jsx(a,{className:"h-full w-8"}),e.jsx("div",{className:"flex flex-row flex-wrap gap-0.5",children:e.jsx(y,{})})]})})]})})})}function y(){const{data:s,isError:r,isPending:t}=m({throwOnError:!0});return r?null:t?e.jsx(a,{className:"w-6"}):e.jsxs("div",{className:"flex h-6 items-center gap-0.5 rounded-md border border-theme-border-moderate px-1 py-0.5 text-text-sm text-theme-text-secondary",children:[e.jsx(u,{className:"h-4 w-4 fill-theme-text-brand"}),s.version]})}function G(){return e.jsxs("div",{children:[e.jsx(N,{}),e.jsxs("div",{className:"layout-container space-y-5 py-5",children:[e.jsx(L,{}),e.jsx(C,{})]})]})}function L(){return e.jsxs(n,{className:"flex flex-col-reverse overflow-hidden lg:flex-row",children:[e.jsxs("div",{className:"w-full px-7 py-5 lg:w-2/3",children:[e.jsx("h2",{className:"text-display-xs font-semibold",children:"Welcome to ZenML"}),e.jsxs("p",{className:"mt-2 text-text-lg text-theme-text-secondary",children:["You successfully installed ZenML dashboard. ",e.jsx("br",{}),"Now you can get started with your new ZenML server.",e.jsx("br",{}),"To get started"," ",e.jsx(v,{className:"link text-theme-text-brand",to:f.onboarding,children:"checkout the onboarding guide"}),"."]}),e.jsx("hr",{className:"mb-5 mt-5 w-[100px]"}),e.jsxs("div",{className:"",children:[e.jsx("p",{className:"mb-1 text-text-sm",children:"Copy your ZenML server URL"}),e.jsx("div",{className:"",children:e.jsx(j,{codeClasses:"truncate",className:"truncate",code:`zenml connect --url=${window.location.origin}`})})]})]}),e.jsx("div",{className:"flex w-full items-center justify-center bg-primary-50 lg:w-1/3",children:e.jsx("img",{alt:"illustration of different purple squares",src:w,className:"h-full w-full"})})]})}function C(){return e.jsxs(n,{className:"flex flex-col-reverse items-center overflow-hidden md:max-h-[200px] md:flex-row",children:[e.jsxs("div",{className:"w-full space-y-3 px-7 py-5 xl:w-4/5",children:[e.jsxs("div",{className:"flex items-center space-x-1",children:[e.jsx(h,{className:"text-text-xs font-semibold",color:"green",children:"NEW"}),e.jsx("h2",{className:"text-text-xl font-semibold",children:"VS Code Quickstart with ZenML"})]}),e.jsxs("p",{className:"text-theme-text-secondary",children:["Get started quickly with ZenML using GitHub Codespaces!",e.jsx("br",{}),"You can run our quickstart guide in a pre-configured environment."]}),e.jsx(p,{size:"sm",className:"w-fit",emphasis:"subtle",asChild:!0,children:e.jsxs("a",{target:"_blank",rel:"noopener noreferrer",href:"https://github.com/zenml-io/zenml",children:["Visit our GitHub repo ",e.jsx(g,{className:"h-5 w-5"})]})})]}),e.jsx("img",{className:"object-contain",src:b,alt:"Gif explaining how to setup codespaces"})]})}export{G as default};
