"""Release [0.55.5].

Revision ID: 0.55.5
Revises: 72675226b2de
Create Date: 2024-03-06 10:31:37.947486

"""

# revision identifiers, used by Alembic.
revision = "0.55.5"
down_revision = "72675226b2de"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
