import{r as c,j as e}from"./@radix-DP6vWzyx.js";import{a7 as x,a8 as d,a9 as m,aa as p,ab as h,ac as j,ad as t,ae as r}from"./index-QQb7wQEC.js";import{C as o}from"./CodeSnippet-CqybNv0k.js";import{I as u}from"./Infobox-BwisKifi.js";const w=c.forwardRef(({name:a,children:i,type:s,...l},n)=>e.jsxs(x,{children:[e.jsx(d,{asChild:!0,children:i}),e.jsxs(m,{className:"w-fit max-w-fit",...l,ref:n,children:[e.jsx(p,{children:e.jsx(h,{children:j(s||"")})}),e.jsxs("div",{className:"space-y-5 p-7",children:[e.jsx(C,{type:s}),e.jsxs("div",{className:"space-y-1",children:[e.jsxs("p",{className:"text-text-sm text-theme-text-secondary",children:["Describe your ",t(s)]}),e.jsx(o,{codeClasses:"whitespace-pre-wrap",wrap:!0,code:`zenml ${r(s)} describe ${a}`})]}),e.jsxs("div",{className:"space-y-1",children:[e.jsxs("p",{className:"text-text-sm text-theme-text-secondary",children:["Update your ",t(s)]}),e.jsx(o,{codeClasses:"whitespace-pre-wrap",wrap:!0,code:`zenml ${r(s)} update ${a}`})]})]})]})]}));w.displayName="ComponentFallbackDialog";function C({type:a}){return e.jsx(u,{children:e.jsx("div",{className:"flex w-full flex-wrap justify-between gap-2",children:e.jsxs("div",{className:"min-w-0",children:[e.jsx("p",{className:"truncate text-text-sm font-semibold",children:"We are working on the new Stack Components experience."}),e.jsxs("p",{className:"truncate text-text-sm",children:["Meanwhile you can use the CLI to manage your ",t(a),"."]})]})})})}export{w as C};
