import{j as e}from"./@radix-DP6vWzyx.js";import{B as d,aQ as r}from"./index-QQb7wQEC.js";const s=["Managed ZenML server on your VPC or hosted on our servers","Social SSO, RBAC, and User Management"],m=["Model Control Plane Dashboard",...s,"CI/CD/CT, Artifact Control Plane and more!"],x=["Artifact Control Plane Dashboard",...s,"CI/CD/CT, Model Control Plane and more!"];function p({features:t,image:a,feature:n}){return e.jsxs(d,{className:"relative overflow-hidden px-7 py-5",children:[e.jsxs("div",{className:"max-w-[450px] space-y-4",children:[e.jsxs("p",{className:"text-display-xs font-semibold",children:["Access Advanced ",n==="artifact"?"Artifact":"Model"," Management Features with ZenML Pro"]}),e.jsx("ul",{className:"space-y-2",children:t.map((o,l)=>e.jsx("li",{className:"text-text-md text-theme-text-secondary",children:o},l))}),e.jsxs("div",{className:"flex flex-wrap items-center gap-2",children:[e.jsx("a",{href:"https://cloud.zenml.io/signup",target:"_blank",rel:"noopener noreferrer",className:r({intent:"primary",size:"md"}),children:"Upgrade to ZenML Pro"}),e.jsx("a",{href:"https://www.zenml.io/cloud",target:"_blank",rel:"noopener noreferrer",className:r({intent:"primary",size:"md",emphasis:"minimal"}),children:"Learn more"})]})]}),e.jsx("img",{className:"absolute right-0 top-0 hidden translate-x-[45%] translate-y-[10%] rounded-md border border-theme-border-moderate lg:block",src:a.src,alt:a.alt})]})}export{p as C,x as a,m};
