import{j as e}from"./@radix-DP6vWzyx.js";import{B as o}from"./index-QQb7wQEC.js";import{a as s}from"./ProviderRadio-BROY1700.js";function m({children:r}){return e.jsx(o,{className:"w-full",children:r})}function u({children:r}){return e.jsx("div",{className:"border-b border-theme-border-moderate px-5 py-3 text-display-xs font-semibold",children:r})}function x({children:r}){return e.jsx("div",{className:"p-5",children:r})}function a({children:r}){return e.jsx("div",{className:"flex items-center justify-end gap-2 border-t border-theme-border-moderate p-5",children:r})}function c({children:r,displayCancel:t=!0}){return e.jsxs(a,{children:[t&&e.jsx(s,{}),r]})}export{x as B,u as H,c as S,m as W};
