"""Release [0.60.0].

Revision ID: 0.60.0
Revises: 25155145c545
Create Date: 2024-06-25 14:26:48.867005

"""

# revision identifiers, used by Alembic.
revision = "0.60.0"
down_revision = "25155145c545"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass
